"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
require("reflect-metadata");
const utils_1 = require("power-di/utils");
const generatorFuncPrototype = Object.getPrototypeOf(function* () { });
function isGeneratorFunction(fn) {
    return typeof fn === 'function' && Object.getPrototypeOf(fn) === generatorFuncPrototype;
}
exports.isGeneratorFunction = isGeneratorFunction;
const asyncFuncPrototype = Object.getPrototypeOf(function () {
    return tslib_1.__awaiter(this, void 0, void 0, function* () { });
});
function isAsyncFunction(fn) {
    return typeof fn === 'function' && Object.getPrototypeOf(fn) === asyncFuncPrototype;
}
exports.isAsyncFunction = isAsyncFunction;
const routes = [];
const methods = ['get', 'put', 'post', 'delete', 'patch'];
function getNameAndMethod(functionName) {
    let name = functionName, functionMethod = 'all';
    functionName = functionName.toLowerCase();
    for (let i = 0; i < methods.length; i++) {
        const method = methods[i];
        if (functionName.startsWith(method)) {
            name = functionName.substring(3);
            functionMethod = method;
            break;
        }
    }
    return { name, method: functionMethod };
}
const COMMENTS = /((\/\/.*$)|(\/\*[\s\S]*?\*\/))/mg;
const DEFAULT_PARAMS = /=[^,]+/mg;
const FAT_ARROWS = /=>.*$/mg;
// TS_TO_ES6
const ASYNC_DEFAULT_PARAMS = ', void 0, void 0, function* (';
function getParameterNames(fn) {
    let code = fn.toString()
        .replace(ASYNC_DEFAULT_PARAMS, '')
        .replace(COMMENTS, '')
        .replace(FAT_ARROWS, '')
        .replace(DEFAULT_PARAMS, '');
    let result = code
        .slice(code.indexOf('(') + 1, code.indexOf(')'))
        .match(/([^\s,]+)/g);
    return result === null
        ? []
        : result;
}
const extRules = {};
const getRuleKey = (target, key) => `${utils_1.getGlobalType(target.constructor)}_${key}`;
function getMethodRules(target, key) {
    const ruleKey = getRuleKey(target, key);
    if (!extRules[ruleKey]) {
        extRules[ruleKey] = {
            param: {},
        };
    }
    return extRules[ruleKey];
}
// #region ParameterDecorator
function FromCustom(custom) {
    return (target, key, index) => {
        const methodRule = getMethodRules(target, key);
        methodRule.param[index] = custom;
    };
}
exports.FromCustom = FromCustom;
function FromBody(paramName) {
    return FromCustom((ctx, name) => ctx.request.body[paramName || name]);
}
exports.FromBody = FromBody;
function FromParam(paramName) {
    return FromCustom((ctx, name) => ctx.params[paramName || name]);
}
exports.FromParam = FromParam;
function FromQuery(paramName) {
    return FromCustom((ctx, name) => ctx.query[paramName || name]);
}
exports.FromQuery = FromQuery;
// #endregion
// #region Controller Middleware
const beforeMiddlewares = {};
const afterMiddlewares = {};
function beforeMiddware(middwares) {
    middwares = [].concat(middwares);
    return (target, key) => {
        const typeGlobalName = utils_1.getGlobalType(target.constructor);
        const route = routes.find(r => r.typeGlobalName === typeGlobalName && r.functionName === key);
        if (route) {
            route.beforeMiddleware.concat(middwares);
        }
        else {
            beforeMiddlewares[getRuleKey(target, key)] = middwares;
        }
    };
}
exports.beforeMiddware = beforeMiddware;
function afterMiddware(middwares) {
    middwares = [].concat(middwares);
    return (target, key) => {
        const typeGlobalName = utils_1.getGlobalType(target.constructor);
        const route = routes.find(r => r.typeGlobalName === typeGlobalName && r.functionName === key);
        if (route) {
            route.afterMiddleware.concat(middwares);
        }
        else {
            afterMiddlewares[getRuleKey(target, key)] = middwares;
        }
    };
}
exports.afterMiddware = afterMiddware;
// #endregion
function routerMetadata(data = {}) {
    return function (target, key) {
        const typeGlobalName = utils_1.getGlobalType(target.constructor);
        const CtrlType = target.constructor;
        const routerFn = target[key];
        const paramTypes = Reflect.getMetadata('design:paramtypes', target, key) || [];
        const typeInfo = Object.assign({ onError: function (ctx, err) {
                ctx.throw(err);
            } }, data, { typeGlobalName, typeClass: CtrlType, functionName: key, paramTypes: getParameterNames(routerFn).map((name, i) => {
                return {
                    name,
                    type: paramTypes[i],
                };
            }), returnType: Reflect.getMetadata('design:returntype', target, key), beforeMiddleware: (data.beforeMiddleware || []).concat(beforeMiddlewares[getRuleKey(target, key)] || []), afterMiddleware: (data.afterMiddleware || []).concat(afterMiddlewares[getRuleKey(target, key)] || []), call: () => target[key] });
        let name = getNameAndMethod(typeInfo.functionName);
        if (!typeInfo.url) {
            const ctrl = typeGlobalName
                .split('_')[0]
                .toLowerCase()
                .replace('controller', '');
            typeInfo.url = `/${ctrl}/${name.name}`;
        }
        if (!typeInfo.method) {
            typeInfo.method = name.method;
        }
        routes.push(typeInfo);
        const methodRules = getMethodRules(target, key);
        const getArgs = (ctx) => {
            return typeInfo.paramTypes.map((p, i) => {
                const name = p.name;
                let argValue = undefined;
                if (methodRules.param[i]) {
                    argValue = methodRules.param[i](ctx, name);
                }
                else {
                    const param = ctx.params || {};
                    const query = ctx.query || {};
                    const body = (ctx.request || {}).body || {};
                    if (name in param) {
                        argValue = param[name];
                    }
                    else if (name in query) {
                        argValue = query[name];
                    }
                    else if (name in body) {
                        argValue = body[name];
                    }
                }
                if (argValue === undefined) {
                    return argValue;
                }
                switch (p.type) {
                    case Number:
                        return parseFloat(argValue);
                    case String:
                        return `${argValue}`;
                    // TODO custom TypeFormatter
                    default:
                        return argValue;
                }
            });
        };
        let value = routerFn;
        if (isGeneratorFunction(routerFn)) {
            value = function* (ctx) {
                // 'this' maybe is Controller or Context, in Chair.
                ctx = (this.request && this.response ? this : this.ctx) || ctx;
                const ctrl = new CtrlType(ctx);
                const args = getArgs(ctx);
                try {
                    const ret = yield routerFn.apply(ctrl, args);
                    if (ret !== undefined) {
                        ctx.body = ret;
                    }
                    return ret;
                }
                catch (error) {
                    typeInfo.onError(ctx, error);
                }
            };
        }
        else {
            value = function (ctx) {
                return tslib_1.__awaiter(this, void 0, void 0, function* () {
                    // 'this' maybe is Controller or Context, in Chair.
                    ctx = (this.request && this.response ? this : this.ctx) || ctx;
                    const ctrl = new CtrlType(ctx);
                    const args = getArgs(ctx);
                    try {
                        const ret = yield Promise.resolve(routerFn.apply(ctrl, args));
                        if (ret !== undefined) {
                            ctx.body = ret;
                        }
                        return ret;
                    }
                    catch (error) {
                        typeInfo.onError(ctx, error);
                    }
                });
            };
        }
        return {
            value,
        };
    };
}
exports.routerMetadata = routerMetadata;
function getRouters() {
    return routes;
}
exports.getRouters = getRouters;
